# Tested under Julia version 1.4.1
# Required dependencies will be automaticall downloaded and installed if needed
# and they are specified in Project.toml and Manifest.toml files.

if length(ARGS) == 1 && ARGS[1] == "-n"
    const sep = 0.0
elseif length(ARGS) == 1 && ARGS[1] == "-s"
    const sep = 0.0450534136444119
else
    println("usage: julia solver.jl [-s|-n]\n")
    println("If -s swithch is passed then optimization is performed with separation.")
    println("If -n swithch is passed then optimization is performed without separation.\n")
    println("Make sure you have Project.toml and Manifest.toml files that come with")
    println("this file downloaded and stored in the same directory.")
    exit(1)
end

using Pkg
Pkg.activate(".")
Pkg.instantiate()

using JuMP
using Ipopt

const n = 1
const a = 0.07
const ε₀ = 2.0^(-32) # some constant small enough

const x0 = 2.0^(-32)
const poly_a = 1.0 / (2.0*x0)
const poly_b = log(x0)
const poly_c = -x0/2.0

# WARNING!
# make sure we use xlog only in objective function and only when subtracting it
# as we use relaxation from above
xlog(x) = x < x0 ? (poly_a*x*x+poly_b*x+poly_c) : x*log(x)
dxlog(x) = x < x0 ? (2.0*poly_a*x+poly_b) : log(x)+1.0
ddxlog(x) = x < x0 ? 2.0*poly_a : 1.0/x

H1(p) = -(xlog(p)+xlog(1-p))
H3(p, q, r) = -(xlog(p)+xlog(q)+xlog(r))
H4(p, q, r, s) = -(xlog(p)+xlog(q)+xlog(r)+xlog(s))
H8(a1,a2,a3,a4,a5,a6,a7,a8) = -(xlog(a1)+xlog(a2)+xlog(a3)+xlog(a4)+xlog(a5)+xlog(a6)+xlog(a7)+xlog(a8))
H6(a1,a2,a3,a4,a5,a6) = -(xlog(a1)+xlog(a2)+xlog(a3)+xlog(a4)+xlog(a5)+xlog(a6))

fun(x) = (sin(pi*(x-0.5))+1.0) / 2.0

dl(x)=x*(1-exp(-x))/(1-exp(-x)-x*exp(-x))

model = Model(optimizer_with_attributes(Ipopt.Optimizer, "max_iter" => 100000, "print_level" => 1))

register(model, :fun, 1, fun, autodiff=true)
register(model, :xlog, 1, xlog, dxlog, ddxlog)
register(model, :H1, 1, H1, autodiff=true)
register(model, :H3, 3, H3, autodiff=true)
register(model, :H4, 4, H4, autodiff=true)
register(model, :H8, 8, H8, autodiff=true)
register(model, :H6, 6, H6, autodiff=true)
register(model, :dl, 1, dl, autodiff=true)

@variable(model, sep <= as <= 1.0-sep, start=rand())
@variable(model, sep <= at <= 1.0-sep, start=rand())
@variable(model, sep <= au <= 1.0-sep, start=rand())
@variable(model, sep <= ar <= 1.0-sep, start=rand())
@variable(model, sep <= gs <= 1.0-sep, start=rand())
@variable(model, sep <= gt <= 1.0-sep, start=rand())
@variable(model, sep <= gu <= 1.0-sep, start=rand())
@variable(model, sep <= gr <= 1.0-sep, start=rand())
@variable(model, sep <= bs <= 1.0-sep, start=rand())
@variable(model, sep <= bt <= 1.0-sep, start=rand())
@variable(model, sep <= bu <= 1.0-sep, start=rand())
@variable(model, sep <= br <= 1.0-sep, start=rand())
@variable(model, sep <= bst <= 1.0-sep, start=rand())
@variable(model, sep <= bsu <= 1.0-sep, start=rand())
@variable(model, sep <= bsr <= 1.0-sep, start=rand())
@variable(model, sep <= btt <= 1.0-sep, start=rand())
@variable(model, sep <= bts <= 1.0-sep, start=rand())
@variable(model, sep <= btu <= 1.0-sep, start=rand())
@variable(model, sep <= buu <= 1.0-sep, start=rand())
@variable(model, sep <= bus <= 1.0-sep, start=rand())
@variable(model, sep <= but <= 1.0-sep, start=rand())
@variable(model, sep <= bur <= 1.0-sep, start=rand())
@variable(model, sep <= brr <= 1.0-sep, start=rand())
@variable(model, sep <= brs <= 1.0-sep, start=rand())
@variable(model, sep <= bru <= 1.0-sep, start=rand())
@variable(model, sep <= gst <= 1.0-sep, start=rand())
@variable(model, sep <= gsu <= 1.0-sep, start=rand())
@variable(model, sep <= gsr <= 1.0-sep, start=rand())
@variable(model, sep <= gtt <= 1.0-sep, start=rand())
@variable(model, sep <= gts <= 1.0-sep, start=rand())
@variable(model, sep <= gtu <= 1.0-sep, start=rand())
@variable(model, sep <= guu <= 1.0-sep, start=rand())
@variable(model, sep <= gus <= 1.0-sep, start=rand())
@variable(model, sep <= gut <= 1.0-sep, start=rand())
@variable(model, sep <= gur <= 1.0-sep, start=rand())
@variable(model, sep <= grr <= 1.0-sep, start=rand())
@variable(model, sep <= grs <= 1.0-sep, start=rand())
@variable(model, sep <= gru <= 1.0-sep, start=rand())
@variable(model, sep <= rst <= 1.0-sep, start=rand())
@variable(model, sep <= rsu <= 1.0-sep, start=rand())
@variable(model, sep <= rsr <= 1.0-sep, start=rand())
@variable(model, sep <= rtt <= 1.0-sep, start=rand())
@variable(model, sep <= rts <= 1.0-sep, start=rand())
@variable(model, sep <= rtu <= 1.0-sep, start=rand())
@variable(model, sep <= ruu <= 1.0-sep, start=rand())
@variable(model, sep <= rus <= 1.0-sep, start=rand())
@variable(model, sep <= rut <= 1.0-sep, start=rand())
@variable(model, sep <= rur <= 1.0-sep, start=rand())
@variable(model, sep <= rrr <= 1.0-sep, start=rand())
@variable(model, sep <= rrs <= 1.0-sep, start=rand())
@variable(model, sep <= rru <= 1.0-sep, start=rand())
@variable(model, sep <= x <= 1.0-sep, start=rand())
@variable(model, sep <= y <= 1.0-sep, start=rand())
@variable(model, sep <= z <= 1.0-sep, start=rand())

@variable(model, ls >= sep, start=rand())
@variable(model, lt >= sep, start=rand())
@variable(model, lu >= sep, start=rand())
@variable(model, lr >= sep, start=rand())

@NLobjective(model, Max,
             H4(fun(as), fun(at), fun(au), fun(ar)) + fun(as)*H1(fun(gs)) +
             fun(at)*H1(fun(gt)) +fun(au)*H1(fun(gu)) + fun(ar)*H1(fun(gr)) +
             2*fun(as)*(H6(fun(bsu), fun(bst), fun(bsr), fun(gsu), fun(gst), fun(gsr)) +
                        fun(bsu)*log((1-fun(gu))*fun(au)) + fun(bst)*log((1-fun(gt))*fun(at)) +
                        fun(bsr)*log((1-fun(gr))*fun(ar)) + fun(gsu)*log(fun(gu)*fun(au)) +
                        fun(gst)*log(fun(gt)*fun(at)) + fun(gsr)*log(fun(gr)*fun(ar)) ) + # f_S
             fun(gs)*fun(as)*(1+fun(bs))*(H3(fun(rsu), fun(rst), fun(rsr)) + fun(rsu)*log(fun(au)) +
                                          fun(rst)*log(fun(at)) + fun(rsr)*log(fun(ar))) + # g_S
             2*fun(at)*(H6(fun(btu), fun(btt), fun(bts), fun(gtu), fun(gtt), fun(gts)) +
                        fun(btu)*log((1-fun(gu))*fun(au)) + fun(btt)*log((1-fun(gt))*fun(at)) +
                        fun(bts)*log((1-fun(gs))*fun(as))+ fun(gtu)*log(fun(gu)*fun(au)) +
                        fun(gtt)*log(fun(gt)*fun(at)) + fun(gts)*log(fun(gs)*fun(as)) ) + # f_T
             fun(gt)*fun(at)*(1+fun(bt))*(H3(fun(rtu), fun(rtt), fun(rts)) + fun(rtu)*log(fun(au)) +
                                          fun(rtt)*log(fun(at)) + fun(rts)*log(fun(as))) + # g_T
             2*fun(ar)*(H6(fun(brs), fun(bru), fun(brr), fun(grs), fun(gru), fun(grr)) +
                        fun(brs)*log((1-fun(gs))*fun(as)) + fun(bru)*log((1-fun(gu))*fun(au)) +
                        fun(brr)*log((1-fun(gr))*fun(ar))+fun(grs)*log(fun(gs)*fun(as)) +
                        fun(gru)*log(fun(gu)*fun(au)) + fun(grr)*log(fun(gr)*fun(ar)) ) + # f_R
             fun(gr)*fun(ar)*(1+fun(br))*(H3(fun(rrs), fun(rru), fun(rrr)) + fun(rrs)*log(fun(as)) +
                                          fun(rru)*log(fun(au)) + fun(rrr)*log(fun(ar))) + # g_R
             2*fun(au)*(H8(fun(bus), fun(bur), fun(but), fun(buu), fun(gus), fun(gur), fun(gut), fun(guu)) +
                        fun(bus)*log((1-fun(gs))*fun(as)) + fun(bur)*log((1-fun(gr))*fun(ar)) +
                        fun(but)*log((1-fun(gt))*fun(at)) + fun(buu)*log((1-fun(gu))*fun(au)) +
                        fun(gus)*log(fun(gs)*fun(as)) + fun(gur)*log(fun(gr)*fun(ar)) +
                        fun(gut)*log(fun(gt)*fun(at)) + fun(guu)*log(fun(gu)*fun(au)) ) + # f_U
             fun(gu)*fun(au)*(1+fun(bu))*(H4(fun(rus), fun(rur), fun(rut), fun(ruu)) +
                                          fun(rus)*log(fun(as)) + fun(rur)*log(fun(ar)) +
                                          fun(rut)*log(fun(at)) + fun(ruu)*log(fun(au))) + # g_U
             fun(as)*(1-fun(gs))*(ls - dl(ls) + dl(ls) * log(dl(ls) / ls) + log(1 - exp(-ls)-ls*exp(-ls))) + # w_S
             fun(at)*(1-fun(gt))*(lt - dl(lt) + dl(lt) * log(dl(lt) / lt) + log(1 - exp(-lt)-lt*exp(-lt))) + # w_T
             fun(au)*(1-fun(gu))*(lu - dl(lu) + dl(lu) * log(dl(lu) / lu) + log(1 - exp(-lu)-lu*exp(-lu))) + # w_U
             fun(ar)*(1-fun(gr))*(lr - dl(lr) + dl(lr) * log(dl(lr) / lr) + log(1 - exp(-lr)-lr*exp(-lr))) + # w_R
             fun(as)*fun(gs)*(-1+fun(bs)-xlog(fun(bs))) + # \tilde w_S
             fun(at)*fun(gt)*(-1+fun(bt)-xlog(fun(bt))) + # \tilde w_T
             fun(au)*fun(gu)*(-1+fun(bu)-xlog(fun(bu))) + # \tilde w_U
             fun(ar)*fun(gr)*(-1+fun(br)-xlog(fun(br))) + # \tilde w_R
             a*log(a) + fun(x)*a*log(fun(au)^2+2*fun(au)*(1-fun(au))+fun(ar)^2) -
             xlog(fun(x)*a) +
             fun(y)*a*log(fun(at)^2) -
             xlog(fun(y)*a) +
             fun(z)*a*log(2*fun(as)*(fun(at)+fun(ar))) -
             xlog(fun(z)*a) # h
            )

@NLconstraint(model, fun(ar) <= 0.995)

# property (5) (alpha)
@NLconstraint(model, fun(as) + fun(at) + fun(au) + fun(ar) == 1)

# property (6) (eq1)
@NLconstraint(model, -ε₀ <= fun(gs)*fun(as) + fun(gt)*fun(at) + fun(gu)*fun(au) +
                            fun(gr)*fun(ar) - 3*exp(-2) <= ε₀)

#property (7) (eq2)
@NLconstraint(model, -ε₀ <= fun(bs)*fun(gs)*fun(as) + fun(bt)*fun(gt)*fun(at) +
                            fun(bu)*fun(gu)*fun(au) + fun(br)*fun(gr)*fun(ar) - exp(-2) <= ε₀)

# property (13)
@NLconstraint(model, 2*fun(btt)*fun(at) + 2*fun(at)*fun(gtt) +
                     fun(gt)*fun(rtt)*fun(at)*(1+fun(bt)) + fun(y)*a <= fun(at) + ε₀)
@NLconstraint(model, 2*fun(btt)*fun(at) + 2*fun(at)*fun(gtt) +
                     fun(gt)*fun(rtt)*fun(at)*(1+fun(bt)) + fun(y)*a <= fun(at) * 2)

# property (12)
@NLconstraint(model, fun(as) >= fun(au))

# property (15)
@NLconstraint(model, 2*fun(au) + fun(gu)*fun(au)*(1+fun(bu)) + 2*fun(as)*(fun(bsu)+fun(gsu)) +
                     fun(gs)*fun(as)*fun(rsu)*(1+fun(bs)) + 2*fun(at)*(fun(btu)+fun(gtu)) +
                     fun(gt)*fun(at)*fun(rtu)*(1+fun(bt)) +
                     2*fun(ar)*(fun(bru)+fun(brr)+fun(gru)+fun(grr)) +
                     fun(gr)*fun(ar)*(fun(rru)+fun(rrr))*(1+fun(br)) + fun(x)*a >=
                     4*exp(-2) + a + 4*fun(au) + fun(at) + 2*fun(ar))

# property (9) (sum1)
@NLconstraint(model, fun(bst) + fun(bsu) + fun(bsr) + fun(gst) + fun(gsu) + fun(gsr) == 1)
@NLconstraint(model, fun(bts) + fun(btt) + fun(btu) + fun(gts) + fun(gtt) + fun(gtu) == 1)
@NLconstraint(model, fun(bus) + fun(but) + fun(buu) + fun(bur) + fun(gus) + fun(gut) + fun(guu) + fun(gur) == 1)
@NLconstraint(model, fun(brs) + fun(bru) + fun(brr) +  fun(grs) + fun(gru) + fun(grr) == 1)

# property (10)
@NLconstraint(model, fun(rst) + fun(rsu) + fun(rsr) == 1)
@NLconstraint(model, fun(rts) + fun(rtt) + fun(rtu) == 1)
@NLconstraint(model, fun(rus) + fun(rut) + fun(ruu) + fun(rur) == 1)
@NLconstraint(model, fun(rrs) + fun(rru) + fun(rrr) == 1)

# property (8) (eq3)
@NLconstraint(model, -ε₀ <= fun(at)*(fun(gtt)+fun(gts)+fun(gtu))+fun(as)*(fun(gst)+fun(gsu)+fun(gsr))+
                            fun(au)*(fun(guu)+fun(gus)+fun(gur)+fun(gut))+
                            fun(ar)*(fun(grs)+fun(gru)+fun(grr)) - exp(-2) <= ε₀)

# the number of fun(gr)een edges
@NLconstraint(model, 2*fun(at)*fun(gts)+2*fun(au)*fun(gus)+2*fun(ar)*fun(grs) == fun(as)*fun(gs)*(1-fun(bs)))
@NLconstraint(model, 2*fun(as)*fun(gst)+2*fun(at)*fun(gtt)+2*fun(au)*fun(gut) == fun(at)*fun(gt)*(1-fun(bt)))
@NLconstraint(model, 2*fun(as)*fun(gsu)+2*fun(at)*fun(gtu)+2*fun(au)*fun(guu)+2*fun(ar)*fun(gru) == fun(au)*fun(gu)*(1-fun(bu)))
@NLconstraint(model, 2*fun(as)*fun(gsr)+2*fun(au)*fun(gur)+2*fun(ar)*fun(grr) == fun(ar)*fun(gr)*(1-fun(br)))

# the number of yellow edges
@NLconstraint(model, fun(x)+fun(y)+fun(z) == 1.0)

@NLconstraint(model, dl(ls)*(1-fun(gs))*fun(as) == 2*fun(au)*fun(bus) + 2*fun(ar)*fun(brs) + 2*fun(at)*fun(bts))
@NLconstraint(model, dl(lt)*(1-fun(gt))*fun(at) == 2*fun(at)*fun(btt) + 2*fun(as)*fun(bst) + 2*fun(au)*fun(but))
@NLconstraint(model, dl(lu)*(1-fun(gu))*fun(au) == 2*fun(as)*fun(bsu) + 2*fun(at)*fun(btu) + 2*fun(au)*fun(buu) + 2*fun(ar)*fun(bru))
@NLconstraint(model, dl(lr)*(1-fun(gr))*fun(ar) == 2*fun(as)*fun(bsr) + 2*fun(au)*fun(bur) + 2*fun(ar)*fun(brr))

optimize!(model)
println((objective=objective_value(model), status=raw_status(model),
        as = value(as), at = value(at), au = value(au), ar = value(ar),
        gs = value(gs), gt = value(gt), gu = value(gu), gr = value(gr),
        bs = value(bs), bt = value(bt), bu = value(bu), br = value(br),
        bst = value(bst), bsu = value(bsu), bsr = value(bsr), btt = value(btt),
        bts = value(bts), btu = value(btu), buu = value(buu), bus = value(bus), but = value(but),
        bur = value(bur), brr = value(brr), brs = value(brs), bru = value(bru),
        gst = value(gst), gsu = value(gsu), gsr = value(gsr), gtt = value(gtt), gts = value(gts),
        gtu = value(gtu), guu = value(guu), gus = value(gus), gut = value(gut), gur = value(gur),
        grr = value(grr), grs = value(grs), gru = value(gru), rst = value(rst),
        rsu = value(rsu), rsr = value(rsr), rtt = value(rtt), rts = value(rts),
        rtu = value(rtu), ruu = value(ruu), rus = value(rus), rut = value(rut),
        rur = value(rur), rrr = value(rrr), rrs = value(rrs), rru = value(rru),
        x = value(x), y = value(y), z = value(z),
        ls = value(ls), lt = value(lt), lu = value(lu), lr = value(lr)))
